package com.cicdi.notify.wechat.subscription;

import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.JSONObject;
import com.alibaba.fastjson.support.hsf.HSFJSONUtils;
import com.cicdi.notify.*;
import com.cicdi.notify.template.TemplateManager;
import com.cicdi.notify.wechat.WechatProvider;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.client.utils.URIBuilder;
import org.apache.http.entity.StringEntity;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.HttpClientBuilder;
import org.apache.http.util.EntityUtils;

import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.time.Duration;
import java.util.Map;
import java.util.concurrent.atomic.AtomicReference;

/**
 * 微信公众号消息通知器
 *
 * @author xueye
 */
public class WechatSubscriptionNotifier extends AbstractNotifier<WechatSubscriptionTemplate> {

    /**
     * 通知器配置信息
     */
    private final NotifierProperties properties;

    /**
     * 全局唯一接口调用凭证
     */
    private final AtomicReference<String> accessToken = new AtomicReference<>();

    /**
     * 获取凭证的Api
     */
    private static final String tokenApi = "https://api.weixin.qq.com/cgi-bin/token";

    /**
     * 发送消息通知的Api
     */
    private static final String notifyApi = " https://api.weixin.qq.com/cgi-bin/message/template/send";

    /**
     * 凭证有效时间
     */
    private static final long tokenTimeOut = Duration.ofSeconds(7000).toMillis();

    /**
     * 凭据刷新时间
     * <p>一般接口对凭证的获取会有次数限制，记录上一次获取凭证的时间，每次发送通知的时候，配合凭证有效时间来判断是否重新获取</p>
     */
    private long refreshTokenTime;

    private final CloseableHttpClient httpClient;

    public WechatSubscriptionNotifier(NotifierProperties properties, TemplateManager templateManager) {
        super(templateManager);
        this.properties = properties;
        this.httpClient = HttpClientBuilder.create().build();
    }

    @Override
    public String getNotifierId() {
        return properties.getId();
    }

    @Override
    public NotifyType getType() {
        return DefaultNotifyType.wechat;
    }

    @Override
    public Provider getProvider() {
//        return WechatProvider.subscription;
        return null;
    }

    @Override
    public void send(WechatSubscriptionTemplate template, Map<String, Object> context) {
        try {
            URI sendMessageURI = new URIBuilder(notifyApi)
                    .addParameter("access_token", getToken())
                    .build();
            HttpPost httpPost = new HttpPost(sendMessageURI);
            httpPost.setHeader("Content-Type", "application/json;charset=utf8");
            StringEntity requestBody = new StringEntity(template.createJsonRequest());
            httpPost.setEntity(requestBody);
            CloseableHttpResponse response = httpClient.execute(httpPost);
            JSONObject responseJson = JSON.parseObject(EntityUtils.toString(response.getEntity()));
            if (!"0".equals(responseJson.get("errcode"))) {
                throw new RuntimeException(responseJson.get("errmsg").toString());
            }
        } catch (URISyntaxException | IOException e) {
            e.printStackTrace();
        }
    }

    @Override
    public void close() {
        accessToken.set(null);
        refreshTokenTime = 0;
        try {
            httpClient.close();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    private String getToken() {
        if (System.currentTimeMillis() - refreshTokenTime > tokenTimeOut || accessToken.get() == null) {
            refreshToken();
        }
        return accessToken.get();
    }

    /**
     * 刷新接口访问凭证
     */
    private void refreshToken() {
        WechatSubscriptionProperties config = new JSONObject(properties.getConfiguration()).toJavaObject(WechatSubscriptionProperties.class);
        try {
            URI getTokenURI = new URIBuilder(tokenApi)
                    .addParameter("grant_type", config.getGrantType())
                    .addParameter("appid", config.getAppId())
                    .addParameter("secret", config.getSecret())
                    .build();
            HttpGet httpGet = new HttpGet(getTokenURI);
            CloseableHttpResponse response = httpClient.execute(httpGet);
            String responseString = EntityUtils.toString(response.getEntity());
            JSONObject responseJson = JSON.parseObject(responseString);
            if (responseJson.containsKey("access_token")) {
                accessToken.set(responseJson.get("access_token").toString());
            } else if (responseJson.containsKey("errcode")) {
                throw new RuntimeException(responseJson.get("errmsg").toString());
            }
        } catch (URISyntaxException | IOException e) {
            e.printStackTrace();
        }
    }

}
